"""Main logic for the `ghab` cli tool."""
import os
import argparse
import math
from typing import Optional, List, Dict

from tqdm import tqdm  # type: ignore

from . import github


def do_argparse() -> argparse.Namespace:
    """Configure an `argparse.ArgumentParser` and use it to parse the command line parameters."""
    parser = argparse.ArgumentParser()
    parser.add_argument("repo", help="The repo to pull worklfow billing info from, eg: nadock/ghab")
    parser.add_argument(
        "--workflow",
        action="append",
        help=(
            "The name of the workflow to pull billing info rom. Can be specified multiple times to pull info for "
            "multiple workflows. If no workflows are specified, billing infor will be pulled for every workflow in the "
            "repository."
        ),
        default=None,
    )
    parser.add_argument(
        "--limit",
        type=int,
        default=100,
        help=(
            "Soft limit on the number of runs per workflow to pull billing info for. Because of the way GitHub's "
            "pagination works, additional records may be retrieved. (default: 100)"
        ),
    )
    parser.add_argument(
        "--github-token",
        help=(
            "The GitHub OAuth token to use in API requests. By default this is pulled from the environment variable "
            "GITHUB_TOKEN."
        ),
        default=None,
    )
    parser.add_argument(
        "--show-runs",
        action="store_true",
        default=False,
        help="Show details of each run, in addition to averages (warning: output can be very long)",
    )
    parser.add_argument(
        "--quiet", action="store_true", default=False, help="Disable all user feedback, just display results.",
    )

    args = parser.parse_args()

    # Get arg.github_token default from OS environment variables
    if args.github_token is None:
        args.github_token = os.environ.get("GITHUB_TOKEN")

    return args


def get_workflow_billing(*, repo: str, limit: int, workflows: List[str], github_token: str, quiet: bool) -> dict:
    """
    Get the details of every run in the requested `workflows` list.

    `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

    `limit` the upper limit on the number of runs to retrieve for each workflow.

    `workflows` must be a list of workflow names for which the billing data of runs will be retrieved.

    `github_token` must be the GitHub OAuth token used to authenticate API requests.
    """
    gh_client = github.GitHubApiClient(github_token=github_token)

    # Get the workflows the user request, or all
    requested_workflows = get_workflows(repo=repo, client=gh_client, workflows=workflows)

    # Get all the runs for each workflow, upto the limited amount for each
    workflow_billing: dict = {}
    for name, id in tqdm(
        requested_workflows.items(),
        desc="Workflow",
        initial=1,
        total=len(requested_workflows.keys()),
        bar_format="{desc} {n_fmt}/{total_fmt}: |{bar}|",
        leave=False,
        disable=quiet,
    ):
        runs = get_workflow_runs(repo=repo, id=id, limit=limit, client=gh_client)

        # For each workflow run, get that runs billing details
        workflow_billing[name] = []
        for run in tqdm(
            runs,
            desc="Workflow runs",
            initial=1,
            bar_format="{desc} {n_fmt}/{total_fmt}: |{bar}|",
            leave=False,
            disable=quiet,
        ):
            billing = get_workflow_run_billing(repo=repo, id=run, client=gh_client)
            workflow_billing[name].append(billing)

    return workflow_billing


def calculate_avg_efficiency(runs: List[Dict[str, Dict[str, float]]]) -> Dict[str, float]:
    """Calculate the average efficiency of all the supplied workflow runs, by OS."""
    ubuntu_total = 0.0
    ubuntu_count = 0
    windows_total = 0.0
    windows_count = 0
    macos_total = 0.0
    macos_count = 0

    for run in runs:
        if "ubuntu" in run and "efficiency" in run["ubuntu"]:
            ubuntu_total += run["ubuntu"]["efficiency"]
            ubuntu_count += 1
        if "windows" in run and "efficiency" in run["windows"]:
            windows_total += run["windows"]["efficiency"]
            windows_count += 1
        if "macos" in run and "efficiency" in run["macos"]:
            macos_total += run["macos"]["efficiency"]
            macos_count += 1

    resp = {}
    if ubuntu_count > 0:
        resp["ubuntu"] = ubuntu_total / ubuntu_count
    if windows_count > 0:
        resp["windows"] = windows_total / windows_count
    if macos_count > 0:
        resp["macos"] = macos_total / macos_count

    return resp


def calculate_avg_billed_ms(runs: List[Dict[str, Dict[str, float]]]) -> Dict[str, float]:
    """Calculate the average billed ms of all the supplied workflow runs, by OS."""
    ubuntu_total = 0.0
    ubuntu_count = 0
    windows_total = 0.0
    windows_count = 0
    macos_total = 0.0
    macos_count = 0

    for run in runs:
        if "ubuntu" in run and "total_ms" in run["ubuntu"]:
            ubuntu_total += run["ubuntu"]["total_ms"]
            ubuntu_count += 1
        if "windows" in run and "total_ms" in run["windows"]:
            windows_total += run["windows"]["total_ms"]
            windows_count += 1
        if "macos" in run and "total_ms" in run["macos"]:
            macos_total += run["macos"]["total_ms"]
            macos_count += 1

    resp = {}
    if ubuntu_count > 0:
        resp["ubuntu"] = ubuntu_total / ubuntu_count
    if windows_count > 0:
        resp["windows"] = windows_total / windows_count
    if macos_count > 0:
        resp["macos"] = macos_total / macos_count

    return resp


def calculate_avg_total_ms(runs: List[Dict[str, int]]) -> Optional[float]:
    """Calculate the average total ms of all the supplied workflow runs, by OS."""
    total = 0
    count = 0

    for run in runs:
        if "total_ms" in run:
            total += run["total_ms"]
            count += 1

    if count == 0:
        return None
    return total / count


def get_workflows(
    *, repo: str, client: github.GitHubApiClient, workflows: Optional[List[str]] = None
) -> Dict[str, int]:
    """
    Get a `dict` of workflow names to the API IDs.

    `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

    `workflows` should be a list of workflow names to retrieve the API IDs for. If it is not specified, the API ID for
    every available workflow in the repository will be retrieved.
    """
    if not client:
        client = github.GitHubApiClient()

    gh_response = client.list_workflows(repo)

    actual_workflows = {}
    for gh_workflow in gh_response:
        if not workflows or gh_workflow["name"] in workflows:
            actual_workflows[gh_workflow["name"]] = gh_workflow["id"]

    return actual_workflows


def get_workflow_runs(*, repo: str, id: int, limit: int, client: github.GitHubApiClient) -> List[int]:
    """
    Get the run IDs for every run of a workflow.

    `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

    `id` must the GitHub API ID of the workflow to retrieve the runs from.

    `limit` the upper limit on the number of runs to retrieve from the workflow.

    `client` should be a `github.GitHubApiClient` instance that will be used to make API requests.
    """
    if not client:
        client = github.GitHubApiClient()

    workflow_runs = client.list_workflow_runs(repo, id, limit=limit)
    return [run["id"] for run in workflow_runs]


def get_workflow_run_billing(*, repo: str, id: int, client: github.GitHubApiClient):
    """
    Get the billing data for a workflow run.

    `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

    `id` must the GitHub API ID of the workflow run of which to retrieve the billing fata.

    `client` should be a `github.GitHubApiClient` instance that will be used to make API requests.
    """
    billing = client.get_workflow_run_usage(repo, id)
    billing_resp = {"total_ms": billing["run_duration_ms"]}

    if "UBUNTU" in billing["billable"]:
        total_ms = billing["billable"]["UBUNTU"]["total_ms"]
        billing_resp["ubuntu"] = {
            "total_ms": total_ms,
            "efficiency": calculate_efficiency(total_ms),
        }
    if "WINDOWS" in billing["billable"]:
        total_ms = billing["billable"]["WINDOWS"]["total_ms"]
        billing_resp["windows"] = {
            "total_ms": total_ms,
            "efficiency": calculate_efficiency(total_ms),
        }
    if "MACOS" in billing["billable"]:
        total_ms = billing["billable"]["MACOS"]["total_ms"]
        billing_resp["macos"] = {
            "total_ms": total_ms,
            "efficiency": calculate_efficiency(total_ms),
        }

    return billing_resp


def calculate_efficiency(total_ms: int):
    """Calculate the efficieny of a workflow from it's total elapsed time in milliseconds."""
    elapsed_minutes = (total_ms / 1000.0) / 60.0
    billed_minutes = math.ceil(elapsed_minutes)
    if billed_minutes == 0:
        return 100.0
    return elapsed_minutes / billed_minutes
