"""API client for interacting with GitHub's V3 REST API."""
import os
from urllib import parse
from typing import Dict, Optional, List, Union

import requests


class GitHubApiClient:
    """
    API client for interacting with GitHub's V3 REST API.

    `github_token` should be a GitHub OAuth token. If it is not supplied, the `GITHUB_TOKEN` os environmnet varibale
    will be used instead. If `github_token` was not supplied and the environment variable is not set a `ValueError` will
    be raised instead.

    `base_url` should the base API url to send requests to, by default it will be `"https://api.github.com"`. However,
    you can override this for unit testing or any other purpose. All requests use `urllib.parse.urljoin` to combine the
    `base_url` with the endpoint specifically for that request.
    """

    def __init__(self, *, github_token: Optional[str] = None, base_url: Optional[str] = None):

        if github_token:
            self.github_token = github_token
        elif os.environ.get("GITHUB_TOKEN"):
            self.github_token = os.environ["GITHUB_TOKEN"]
        else:
            raise ValueError("No GitHub token provided or availble in environment")

        if base_url:
            self.base_url = base_url
        else:
            self.base_url = "https://api.github.com"

        self._session = requests.Session()

    def headers(self, headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """
        Build and return a `dict` of default headers for every request.

        `headers` will be merged into the outputted headers, overriding the defaults if the is a conflict.
        """
        if not headers:
            headers = {}

        return {
            "Authorization": f"Token {self.github_token}",
            "Accept": "application/vnd.github.v3+json",
            **headers,
        }

    def get(
        self, endpoint: str, *, params: Optional[dict] = None, headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        """
        Perform a HTTP GET call to the GitHub API.

        The response body will be checked to ensure it is valid JSON before passing the response back. If the body is
        invalid, a `ValueError` will be raised.

        The status code of the response will be checked to ensure the request was successful before passing the response
        back. If the request was unsucessful, a `ValueError` will be raised.

        `endpoint` must be the GitHub API endpoint that should be joined with the base URL to form the request's full
        URL.

        `params` should be a `dict` of querystring parameters to add to the GET request.

        `headers` should be a `dict` of additional headers to add to the GET request, beyond the defaults from
        `self.headers`.
        """

        url = parse.urljoin(self.base_url, endpoint)
        headers = self.headers(headers=headers)

        response = self._session.get(url, params=params, headers=headers)

        try:
            response.json()
        except Exception as ex:
            raise ValueError(f"GitHub API response body not JSON: {ex}") from ex

        try:
            response.raise_for_status()
        except requests.HTTPError as ex:
            raise ValueError(f"GitHub API request failure HTTP_{response.status_code}: {ex}: {response.json()}") from ex

        return response

    def paginate_response(self, response: requests.Response, list_key: str, limit: Optional[int] = None) -> List[dict]:
        """
        Repeat requests for each paginated page until all values have been retrieved of the number of items retrieved is
        equal to or greater than `limit`.

        `repsonse` must be the response from the first request to a paginated endpoint.

        `list_key` must be the key into the response body that contains the list of paginated items retrieved in each
        request.

        `limit` is an optional upper bound on the number of items retrieved from the API. If the last pagination request
        exceeds this limit, requests will stop and all retrieved values will be returned. Due to the nature of GitHub's
        pagination implementation, large values (> 100) for `limit` can result in more than `limit` items being
        returned.
        """
        # Handle body of first response
        results = response.json()[list_key]

        # Setup second request
        next_page = parse_pagination_header(response.headers.get("link")).get("next")
        if limit and len(results) >= limit:
            next_page = None

        # Get paginated responses
        while next_page:
            response = self.get(next_page)
            results += response.json()[list_key]
            next_page = next_page = parse_pagination_header(response.headers.get("link")).get("next")

            if limit and len(results) >= limit:
                next_page = None

        return results

    def list_workflows(self, repo: str, limit: Optional[int] = None) -> List[dict]:
        """
        List all the GitHub Actions workflows available in the named repository.

        `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

        `limit` is an optional upper bound on the number of items retrieved from the API. If the last pagination request
        exceeds this limit, requests will stop and all retrieved values will be returned. Due to the nature of GitHub's
        pagination implementation, large values (> 100) for `limit` can result in more than `limit` items being
        returned.
        """
        per_page = 100
        if limit and limit < 100:
            per_page = limit

        response = self.get(f"/repos/{repo}/actions/workflows", params={"per_page": per_page})
        return self.paginate_response(response, "workflows")

    def list_workflow_runs(self, repo: str, workflow: Union[str, int], limit: Optional[int] = None) -> List[dict]:
        """
        List all the runs for a specific GitHub Actions workflow available in the named repository.

        `https://developer.github.com/v3/actions/workflow-runs/#list-workflow-runs`

        `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

        `worflow` must be the name or ID of a GitHub Actions workflow available in the named repository.

        `limit` is an optional upper bound on the number of items retrieved from the API. If the last pagination request
        exceeds this limit, requests will stop and all retrieved values will be returned. Due to the nature of GitHub's
        pagination implementation, large values (> 100) for `limit` can result in more than `limit` items being
        returned.
        """
        per_page = 100
        if limit and limit < 100:
            per_page = limit

        response = self.get(f"/repos/{repo}/actions/workflows/{workflow}/runs", params={"per_page": per_page})
        return self.paginate_response(response, "workflow_runs", limit=limit)

    def list_all_workflow_runs(self, repo: str, limit: Optional[int] = None) -> List[dict]:
        """
        List all the GitHub Actions workflow runs available in the named repository, for all workflows.

        `https://developer.github.com/v3/actions/workflow-runs/#list-workflow-runs-for-a-repository`

        `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

        `limit` is an optional upper bound on the number of items retrieved from the API. If the last pagination request
        exceeds this limit, requests will stop and all retrieved values will be returned. Due to the nature of GitHub's
        pagination implementation, large values (> 100) for `limit` can result in more than `limit` items being
        returned.
        """
        per_page = 100
        if limit and limit < 100:
            per_page = limit

        response = self.get(f"/repos/{repo}/actions/runs", params={"per_page": per_page})
        return self.paginate_response(response, "workflow_runs", limit=limit)

    def get_workflow_run_usage(self, repo: str, run: Union[str, int]) -> dict:
        """
        Get the usage information for a GitHub Actions workflow run.

        `https://developer.github.com/v3/actions/workflow-runs/#get-workflow-run-usage`

        `repo` must be the name of the requested repository including the owner section, eg: `nadock/ghab`.

        `run` mut be the name or ID of a GitHub Actions workflow run available in the named repository.
        """
        response = self.get(f"/repos/{repo}/actions/runs/{run}/timing")
        return response.json()


def parse_pagination_header(header: Optional[str]) -> Dict[str, str]:
    """
    Parse a GitHub `Link` pagination header, returing a `dict` containing the name of each pagination link as keys
    and the URLs of the links as values.

    More information about GitHub's `Link` pagination headers can be found in their API docs here:
    `https://developer.github.com/v3/#pagination`.
    """
    links: Dict[str, str] = {}
    if not header:
        return links

    header_links = header.split(", ")
    for link in header_links:
        try:
            url, ref = link.split("; ")
            url = url[1:-1]
            ref = ref[5:-1]
            links[ref] = url
        except IndexError:
            # Ignore invalid pagnation values
            pass

    return links
